require 'spec_helper'
require 'active_record'

describe Hoodoo::ActiveRecord::Dated do

  before :all do
    spec_helper_silence_stdout() do
      ActiveRecord::Migration.create_table( :r_spec_model_effective_date_tests, :id => :string ) do | t |
        t.text :data
        t.timestamps :null => true
      end

      ActiveRecord::Migration.change_column( :r_spec_model_effective_date_tests, :id, :string, :limit => 32 )

      ActiveRecord::Migration.create_table( :r_spec_model_effective_date_tests_history_entries, :id => :string ) do | t |
        t.text     :uuid,            :null => false
        t.text     :data
        t.datetime :effective_start, :null => false
        t.datetime :effective_end,   :null => false

        t.timestamps :null => true
      end

      class RSpecModelEffectiveDateTest < ActiveRecord::Base
        include Hoodoo::ActiveRecord::Dated

        dating_enabled()
      end

      ActiveRecord::Migration.create_table( :r_spec_model_effective_date_test_overrides, :id => :string ) do | t |
        t.text :data
        t.timestamps :null => true
      end

      ActiveRecord::Migration.change_column( :r_spec_model_effective_date_test_overrides, :id, :string, :limit => 32 )

      ActiveRecord::Migration.create_table( :r_spec_model_effective_date_history_entries, :id => :string ) do | t |
        t.text     :uuid,            :null => false
        t.text     :data
        t.datetime :effective_start, :null => false
        t.datetime :effective_end,   :null => false

        t.timestamps :null => true
      end

      class RSpecModelEffectiveDateTestOverride < ActiveRecord::Base
        include Hoodoo::ActiveRecord::Dated
        include Hoodoo::ActiveRecord::Finder

        dating_enabled( :history_table_name => :r_spec_model_effective_date_history_entries )
      end
    end
  end

  shared_examples Hoodoo::ActiveRecord::Dated do

    before( :each ) do
      interaction = Hoodoo::Services::Middleware::Interaction.new( {}, nil )
      interaction.context = Hoodoo::Services::Context.new(
        Hoodoo::Services::Session.new,
        interaction.context.request,
        interaction.context.response,
        interaction
      )

      @context = interaction.context

      # Create some example data for finding. The data has two different UUIDs
      # which I'll referer to as A and B. The following tables contain the
      # historical and current records separately with their attributes, with
      # items created in the historical or main database tables respectively.
      #
      # Historical:
      #
      # -------------------------------------------------------------------
      #  uuid | data    | created_at    | effective_end | effective_start |
      # -------------------------------------------------------------------
      #  A    | "one"   | now - 5 hours | now - 3 hours | now - 5 hours   |
      #  B    | "two"   | now - 4 hours | now - 2 hours | now - 4 hours   |
      #  A    | "three" | now - 5 hours | now - 1 hour  | now - 3 hours   |
      #  B    | "four"  | now - 4 hours | now           | now - 2 hour    |
      #
      # Current:
      #
      # --------------------------------
      #  uuid | data   | created_at    |
      # --------------------------------
      #  B    | "five" | now - 4 hours |
      #  A    | "six"  | now - 5 hours |

      @uuid_a = Hoodoo::UUID.generate
      @uuid_b = Hoodoo::UUID.generate

      @now = Time.now.utc

      # uuid, data, created_at, effective_end, effective_start
      [
        [ @uuid_a, "one",   @now - 5.hours, @now - 3.hours, @now - 5.hours ],
        [ @uuid_b, "two",   @now - 4.hours, @now - 2.hours, @now - 4.hours ],
        [ @uuid_a, "three", @now - 5.hours, @now - 1.hour,  @now - 3.hours ],
        [ @uuid_b, "four",  @now - 4.hours, @now,           @now - 2.hours ]
      ].each do | row_data |
        model_klass.dated_with.new( {
          :id              => row_data[ 0 ] + "-" + row_data[ 3 ].iso8601,
          :uuid            => row_data[ 0 ],
          :data            => row_data[ 1 ],
          :created_at      => row_data[ 2 ],
          :effective_end   => row_data[ 3 ],
          :effective_start => row_data[ 4 ]
        } ).save!
      end

      # uuid, data, created_at, updated_at
      [
        [ @uuid_b, "five", @now - 4.hours, @now          ],
        [ @uuid_a, "six",  @now - 5.hours, @now - 1.hour ]
      ].each do | row_data |
        model_klass.new( {
          :id         => row_data[ 0 ],
          :data       => row_data[ 1 ],
          :created_at => row_data[ 2 ],
          :updated_at => row_data[ 3 ]
        } ).save!
      end

    end

    context 'unscoped' do
      it 'counts only the current records in the main database table' do
        expect( model_klass.count ).to be 2
      end

      it 'finds only the current records in the main database table' do
        expect( model_klass.pluck( :data ) ).to match_array( [ 'five', 'six' ] )
      end
    end

    context '#dating_enabled?' do
      class RSpecNotDated < Hoodoo::ActiveRecord::Base; end

      it 'says it is automatically dated' do
        expect( model_klass.dating_enabled? ).to eq( true )
      end

      it 'knows when something is not automatically dated' do
        expect( RSpecNotDated.dating_enabled? ).to eq( false )
      end
    end

    context '#dated_at' do
      it 'returns counts correctly' do
        expect( model_klass.dated_at( @now - 10.hours ).count ).to be 0
        expect( model_klass.dated_at( @now ).count ).to be 2
      end

      def test_expectation( time, expected_data )
        expect( model_klass.dated_at( time ).pluck( :data ) ).to match_array( expected_data )
      end

      it 'returns no records before any were effective' do
        test_expectation( @now - 10.hours, [] )
      end

      it 'returns records that used to be effective starting at past time' do
        test_expectation( @now - 5.hours, [ "one"           ] )
        test_expectation( @now - 4.hours, [ "one", "two"    ] )
        test_expectation( @now - 3.hours, [ "two", "three"  ] )
        test_expectation( @now - 2.hours, [ "three", "four" ] )
        test_expectation( @now - 1.hour,  [ "four", "six"   ] )
      end

      it 'returns records that are effective now' do
        test_expectation( @now, [ "five", "six" ] )
      end

      it 'works with further filtering' do
        expect( model_klass.dated_at( @now ).where( :id => @uuid_a ).pluck( :data ) ).to eq( [ "six" ] )
      end

    end

    context '#dated' do
      it 'returns counts correctly' do
        @context.request.dated_at = @now - 10.hours
        expect( model_klass.dated( @context ).count ).to be 0

        @context.request.dated_at = @now
        expect( model_klass.dated( @context ).count ).to be 2
      end

      def test_expectation( time, expected_data )
        @context.request.dated_at = time
        expect( model_klass.dated( @context ).pluck( :data ) ).to match_array( expected_data )
      end

      it 'returns no records before any were effective' do
        test_expectation( @now - 10.hours, [] )
      end

      it 'returns records that used to be effective starting at past time' do
        test_expectation( @now - 5.hours, [ "one"           ] )
        test_expectation( @now - 4.hours, [ "one", "two"    ] )
        test_expectation( @now - 3.hours, [ "two", "three"  ] )
        test_expectation( @now - 2.hours, [ "three", "four" ] )
        test_expectation( @now - 1.hour,  [ "four", "six"   ] )
      end

      it 'returns records that are effective now' do
        test_expectation( @now, [ "five", "six" ] )
      end

      it 'works with further filtering' do
        @context.request.dated_at = @now
        expect( model_klass.dated( @context ).where( :id => @uuid_a ).pluck( :data ) ).to eq( [ "six" ] )
      end

      it 'works with dating last' do
        @context.request.dated_at = @now
        expect( model_klass.where( :id => @uuid_a ).dated( @context ).pluck( :data ) ).to eq( [ "six" ] )
      end
    end

    context '#dated_historical_and_current' do
      it 'returns counts correctly' do
        expect( model_klass.dated_historical_and_current.count ).to be 6
      end

      it 'lists all historical and current records' do
        expect( model_klass.dated_historical_and_current.pluck( :data ) ).to match_array( [ 'one', 'two', 'three', 'four', 'five', 'six' ] )
      end

      context 'SQL' do
        it 'has expected default columns' do
          sql = model_klass.dated_historical_and_current.to_sql.downcase

          expect( sql ).to include( 'select "id","data","created_at","updated_at"' )
          expect( sql ).to include( 'select "uuid" as "id","data","created_at","updated_at"' )
        end

        it 'handles custom column selections' do
          sql = model_klass.dated_historical_and_current(
            unquoted_column_names: [ 'id', 'created_at' ]
          ).to_sql.downcase

          expect( sql ).to include( 'select "id","created_at"' )
          expect( sql ).to include( 'select "uuid" as "id","created_at"' )
        end

        it 'handles custom column selections that omit "id"' do
          sql = model_klass.dated_historical_and_current(
            unquoted_column_names: [ 'created_at' ]
          ).to_sql.downcase

          expect( sql ).to include( 'select "created_at","id"' )
          expect( sql ).to include( 'select "created_at","uuid" as "id"' )
        end
      end
    end

    # These appear in the shared examples to take advantage of the test
    # data that's set up, but only work when "model_klass" is the
    # Finder-extended "RSpecModelEffectiveDateTestOverride".
    #
    # Hence "next unless..." in each example (using "skip" in a "before"
    # filter would result in RSpect listing pending tests; they are not
    # pending at all).

    context 'with Hoodoo::ActiveRecord::Finder support' do
      context '#scoped_in' do
        it 'generates appropriate scope' do
          next unless model_klass == RSpecModelEffectiveDateTestOverride

          expect( Hoodoo::ActiveRecord::Support ).to(
            receive( :full_scope_for ).once().with(
              RSpecModelEffectiveDateTestOverride, @context
            ).and_call_original()
          )

          sql = RSpecModelEffectiveDateTestOverride.scoped_in( @context ).to_sql

          expect( sql ).to include( "UNION ALL" )
          expect( sql ).to include( "' AND (\"effective_end\" > '" )
        end

        it 'generates appropriate undated scope' do
          next unless model_klass == RSpecModelEffectiveDateTestOverride

          expect( Hoodoo::ActiveRecord::Support ).to(
            receive( :add_undated_scope_to ).once().with(
              kind_of( ActiveRecord::Relation ), RSpecModelEffectiveDateTestOverride, @context
            ).and_call_original()
          )

          sql = RSpecModelEffectiveDateTestOverride.scoped_undated_in( @context ).to_sql

          expect( sql ).to eq( "SELECT \"r_spec_model_effective_date_test_overrides\".* FROM \"r_spec_model_effective_date_test_overrides\"" )
        end
      end

      context '#acquire_in' do
        context 'with contemporary' do
          it 'finds the contemporary record' do
            next unless model_klass == RSpecModelEffectiveDateTestOverride

            @context.request.dated_at            = nil
            @context.request.uri_path_components = [ @uuid_a ]

            result = RSpecModelEffectiveDateTestOverride.acquire_in!( @context )
            expect( result ).to_not be_nil

            expect( @context.response.halt_processing? ).to eq( false )
            expect( result.data                        ).to eq( 'six' )
          end

          it 'finds a historic record' do
            next unless model_klass == RSpecModelEffectiveDateTestOverride

            @context.request.dated_at            = @now - 4.hours
            @context.request.uri_path_components = [ @uuid_a ]

            result = RSpecModelEffectiveDateTestOverride.acquire_in!( @context )
            expect( result ).to_not be_nil

            expect( @context.response.halt_processing? ).to eq( false )
            expect( result.data                        ).to eq( 'one' )
          end

          it 'indicates correctly that a contemporary exists during a far-backdated lookup' do
            next unless model_klass == RSpecModelEffectiveDateTestOverride

            @context.request.dated_at            = @now - 1.year
            @context.request.uri_path_components = [ @uuid_a ]

            result = RSpecModelEffectiveDateTestOverride.acquire_in!( @context )
            expect( result ).to be_nil

            expect( @context.response.halt_processing?    ).to eq( true )
            expect( @context.response.errors.errors.count ).to eq( 2    )

            expect( @context.response.errors.errors[ 0 ][ 'code'      ] ).to eq( 'generic.not_found' )
            expect( @context.response.errors.errors[ 1 ][ 'code'      ] ).to eq( 'generic.contemporary_exists' )
            expect( @context.response.errors.errors[ 0 ][ 'reference' ] ).to eq( @uuid_a )
            expect( @context.response.errors.errors[ 1 ][ 'reference' ] ).to eq( @uuid_a )
          end
        end

        context 'with a UUID that matches no existing record' do
          it 'indicates correctly that no record exists and does not say a contemporary is present' do
            next unless model_klass == RSpecModelEffectiveDateTestOverride

            alt_uuid                             = Hoodoo::UUID.generate
            @context.request.dated_at            = @now - 5.seconds
            @context.request.uri_path_components = [ alt_uuid ]

            result = RSpecModelEffectiveDateTestOverride.acquire_in!( @context )
            expect( result ).to be_nil

            expect( @context.response.halt_processing?    ).to eq( true )
            expect( @context.response.errors.errors.count ).to eq( 1    )

            expect( @context.response.errors.errors[ 0 ][ 'code'      ] ).to eq( 'generic.not_found' )
            expect( @context.response.errors.errors[ 0 ][ 'reference' ] ).to eq( alt_uuid )
          end
        end

        context 'without contemporary' do
          before( :each ) do
            if model_klass == RSpecModelEffectiveDateTestOverride
              record = RSpecModelEffectiveDateTestOverride.find( @uuid_a )
              record.destroy!

              sleep( 0.1 ) # Be sure that there's nothing left at "now", within timer resolution
            end
          end

          it 'finds no contemporary record' do
            next unless model_klass == RSpecModelEffectiveDateTestOverride

            @context.request.dated_at            = nil
            @context.request.uri_path_components = [ @uuid_a ]

            result = RSpecModelEffectiveDateTestOverride.acquire_in!( @context )
            expect( result ).to be_nil

            expect( @context.response.halt_processing?    ).to eq( true )
            expect( @context.response.errors.errors.count ).to eq( 1    )

            expect( @context.response.errors.errors[ 0 ][ 'code'      ] ).to eq( 'generic.not_found' )
            expect( @context.response.errors.errors[ 0 ][ 'reference' ] ).to eq( @uuid_a )
          end

          it 'finds a historic record' do
            next unless model_klass == RSpecModelEffectiveDateTestOverride

            @context.request.dated_at            = @now - 4.hours
            @context.request.uri_path_components = [ @uuid_a ]

            result = RSpecModelEffectiveDateTestOverride.acquire_in!( @context )
            expect( result ).to_not be_nil

            expect( @context.response.halt_processing? ).to eq( false )
            expect( result.data                        ).to eq( 'one' )
          end

          it 'indicates correctly that no contemporary exists during a far-backdated lookup' do
            next unless model_klass == RSpecModelEffectiveDateTestOverride

            @context.request.dated_at            = @now - 1.year
            @context.request.uri_path_components = [ @uuid_a ]

            result = RSpecModelEffectiveDateTestOverride.acquire_in!( @context )
            expect( result ).to be_nil

            expect( @context.response.halt_processing?    ).to eq( true )
            expect( @context.response.errors.errors.count ).to eq( 1    )

            expect( @context.response.errors.errors[ 0 ][ 'code'      ] ).to eq( 'generic.not_found' )
            expect( @context.response.errors.errors[ 0 ][ 'reference' ] ).to eq( @uuid_a )
          end
        end
      end
    end

  end # 'shared_examples Hoodoo::ActiveRecord::Dated do'

  context "using default effective dating config" do

    # Must be defined as a method rather than using a 'let' statement as
    # 'let' statement values cannot be used in 'before' blocks.
    #
    def model_klass
      RSpecModelEffectiveDateTest
    end

    it_behaves_like Hoodoo::ActiveRecord::Dated

  end

  context "overriding history table name" do

    # Must be defined as a method rather than using a 'let' statement as
    # 'let' statement values cannot be used in 'before' blocks.
    #
    def model_klass
      RSpecModelEffectiveDateTestOverride
    end

    it_behaves_like Hoodoo::ActiveRecord::Dated

  end

  context "SQL and column selections" do
    before :each do
      @now      = Time.now.utc
      @safe_now = RSpecModelEffectiveDateTestOverride.connection.quoted_date( @now )

      request   = Hoodoo::Services::Request.new
      @context  = Hoodoo::Services::Context.new( nil, request, nil, nil )

      @context.request.dated_at = @now
    end

    def run_other_expectations( sql )
      expect( sql ).to include( "from r_spec_model_effective_date_history_entries" )
      expect( sql ).to include( "\"effective_start\" <= '#{ @safe_now }'" )
      expect( sql ).to include( "\"effective_end\" > '#{ @safe_now }'" )
      expect( sql ).to include( "\"effective_end\" is null" )
    end

    it 'generates expected basic SQL' do
      sql = RSpecModelEffectiveDateTestOverride.dated( @context ).to_sql.downcase

      expect( sql ).to include( 'select "id","data","created_at","updated_at"' )
      expect( sql ).to include( 'select "uuid" as "id","data","created_at","updated_at"' )
      run_other_expectations( sql )
    end

    it 'generates expected column-selected SQL via #dated' do
      sql = RSpecModelEffectiveDateTestOverride.dated(
        @context,
        unquoted_column_names: [ 'id', 'created_at' ]
      ).to_sql.downcase

      expect( sql ).to include( 'select "id","created_at"' )
      expect( sql ).to include( 'select "uuid" as "id","created_at"' )
      run_other_expectations( sql )
    end

    it 'includes "id" if omitted, via #dated' do
      sql = RSpecModelEffectiveDateTestOverride.dated(
        @context,
        unquoted_column_names: [ 'created_at' ]
      ).to_sql.downcase

      expect( sql ).to include( 'select "created_at","id"' )
      expect( sql ).to include( 'select "created_at","uuid" as "id"' )
      run_other_expectations( sql )
    end

    it 'generates expected column-selected SQL via #dated_at' do
      sql = RSpecModelEffectiveDateTestOverride.dated_at(
        @now,
        unquoted_column_names: [ 'id', 'created_at' ]
      ).to_sql.downcase

      expect( sql ).to include( 'select "id","created_at"' )
      expect( sql ).to include( 'select "uuid" as "id","created_at"' )
      run_other_expectations( sql )
    end

    it 'includes "id" if omitted, via #dated_at' do
      sql = RSpecModelEffectiveDateTestOverride.dated_at(
        @now,
        unquoted_column_names: [ 'created_at' ]
      ).to_sql.downcase

      expect( sql ).to include( 'select "created_at","id"' )
      expect( sql ).to include( 'select "created_at","uuid" as "id"' )
      run_other_expectations( sql )
    end
  end
end
